"""classicML中的benchmarks用于评估和测试模型的性能和开销"""
from classicML.benchmarks.wrapper_utils import average_timer
from classicML.benchmarks.wrapper_utils import memory_monitor
from classicML.benchmarks.wrapper_utils import timer
