"""tree是classicML中关于树操作的模块, 它可以实现树的生成, 剪枝等功能."""
from classicML.backend.python.tree import criteria
from classicML.backend.python.tree import generators
from classicML.backend.python.tree import pruners